# -*- coding: utf-8 -*-

from nw.gui.about import GuiAbout
from nw.gui.build import GuiBuildNovel
from nw.gui.doceditor import GuiDocEditor
from nw.gui.docmerge import GuiDocMerge
from nw.gui.docsplit import GuiDocSplit
from nw.gui.docviewer import GuiDocViewer, GuiDocViewDetails
from nw.gui.itemdetails import GuiItemDetails
from nw.gui.itemeditor import GuiItemEditor
from nw.gui.mainmenu import GuiMainMenu
from nw.gui.outline import GuiOutline
from nw.gui.outlinedetails import GuiOutlineDetails
from nw.gui.preferences import GuiPreferences
from nw.gui.projload import GuiProjectLoad
from nw.gui.projsettings import GuiProjectSettings
from nw.gui.projtree import GuiProjectTree
from nw.gui.projwizard import GuiProjectWizard
from nw.gui.statusbar import GuiMainStatus
from nw.gui.theme import GuiTheme
from nw.gui.writingstats import GuiWritingStats

__all__ = [
    "GuiAbout",
    "GuiBuildNovel",
    "GuiDocEditor",
    "GuiDocMerge",
    "GuiDocSplit",
    "GuiDocViewDetails",
    "GuiDocViewer",
    "GuiItemDetails",
    "GuiItemEditor",
    "GuiMainMenu",
    "GuiMainStatus",
    "GuiOutline",
    "GuiOutlineDetails",
    "GuiPreferences",
    "GuiProjectLoad",
    "GuiProjectSettings",
    "GuiProjectTree",
    "GuiProjectWizard",
    "GuiTheme",
    "GuiWritingStats",
]
